# coding: utf-8


import os
import uuid
import yaml

from enum import Enum


class ModelModeEnum(Enum):
    """
    Rekcurd model storage options.
    """
    LOCAL = 'local'
    CEPH_S3 = 'ceph_s3'
    AWS_S3 = 'aws_s3'
    GCS = 'gcs'

    @classmethod
    def to_Enum(cls, mode: str):
        if cls.LOCAL.value == mode:
            return cls.LOCAL
        elif cls.CEPH_S3.value == mode:
            return cls.CEPH_S3
        elif cls.AWS_S3.value == mode:
            return cls.AWS_S3
        elif cls.GCS.value == mode:
            return cls.GCS
        else:
            raise ValueError("'{}' is not supported as ModelModeEnum".format(mode))


class RekcurdConfig:
    """
    Rekcurd configurations.
    """
    __SERVICE_DEFAULT_HOST = "127.0.0.1"
    __SERVICE_DEFAULT_PORT = 5000
    __CEPH_DEFAULT_PORT = 8773
    KUBERNETES_MODE: str = None
    DEBUG_MODE: bool = None
    APPLICATION_NAME: str = None
    SERVICE_ID: str = None
    SERVICE_INSECURE_HOST: str = __SERVICE_DEFAULT_HOST
    SERVICE_INSECURE_PORT: int = __SERVICE_DEFAULT_PORT
    SERVICE_LEVEL: str = None
    MODEL_MODE_ENUM: ModelModeEnum = None
    MODEL_FILE_PATH: str = None
    CEPH_ACCESS_KEY: str = None
    CEPH_SECRET_KEY: str = None
    CEPH_HOST: str = None
    CEPH_PORT: int = __CEPH_DEFAULT_PORT
    CEPH_IS_SECURE: bool = None
    CEPH_BUCKET_NAME: str = None
    AWS_ACCESS_KEY: str = None
    AWS_SECRET_KEY: str = None
    AWS_BUCKET_NAME: str = None
    GCS_ACCESS_KEY: str = None
    GCS_SECRET_KEY: str = None
    GCS_BUCKET_NAME: str = None

    def __init__(self, config_file: str = None):
        self.KUBERNETES_MODE = os.getenv("REKCURD_KUBERNETES_MODE")
        if self.KUBERNETES_MODE is None:
            self.__load_from_file(config_file)
        else:
            self.__load_from_env()

    def set_configurations(
            self, debug_mode: bool = None, application_name: str = None,
            service_insecure_host: str = None, service_insecure_port: int = None,
            service_level: str = None, model_mode: str = None,
            model_filepath: str = None, ceph_access_key: str = None,
            ceph_secret_key: str = None, ceph_host: str = None,
            ceph_port: int = None, ceph_is_secure: bool = None,
            ceph_bucket_name: str = None, aws_access_key: str = None,
            aws_secret_key: str = None, aws_bucket_name: str = None,
            gcs_access_key: str = None, gcs_secret_key: str = None, gcs_bucket_name: str = None,
            **options):
        self.DEBUG_MODE = debug_mode if debug_mode is not None else self.DEBUG_MODE
        self.APPLICATION_NAME = application_name or self.APPLICATION_NAME
        self.SERVICE_INSECURE_HOST = service_insecure_host or self.SERVICE_INSECURE_HOST
        self.SERVICE_INSECURE_PORT = int(service_insecure_port or self.SERVICE_INSECURE_PORT)
        self.SERVICE_LEVEL = service_level or self.SERVICE_LEVEL
        if model_mode is not None:
            self.MODEL_MODE_ENUM = ModelModeEnum.to_Enum(model_mode)
        self.MODEL_FILE_PATH = model_filepath or self.MODEL_FILE_PATH
        self.CEPH_ACCESS_KEY = ceph_access_key or self.CEPH_ACCESS_KEY
        self.CEPH_SECRET_KEY = ceph_secret_key or self.CEPH_SECRET_KEY
        self.CEPH_HOST = ceph_host or self.CEPH_HOST
        self.CEPH_PORT = int(ceph_port or self.CEPH_PORT)
        self.CEPH_IS_SECURE = ceph_is_secure if ceph_is_secure is not None else self.CEPH_IS_SECURE
        self.CEPH_BUCKET_NAME = ceph_bucket_name or self.CEPH_BUCKET_NAME
        self.AWS_ACCESS_KEY = aws_access_key or self.AWS_ACCESS_KEY
        self.AWS_SECRET_KEY = aws_secret_key or self.AWS_SECRET_KEY
        self.AWS_BUCKET_NAME = aws_bucket_name or self.AWS_BUCKET_NAME
        self.GCS_ACCESS_KEY = gcs_access_key or self.GCS_ACCESS_KEY
        self.GCS_SECRET_KEY = gcs_secret_key or self.GCS_SECRET_KEY
        self.GCS_BUCKET_NAME = gcs_bucket_name or self.GCS_BUCKET_NAME

    def __load_from_file(self, config_file: str):
        if config_file is not None:
            with open(config_file, 'r') as f:
                config = yaml.load(f)
        else:
            config = dict()
        self.DEBUG_MODE = config.get("debug", True)
        config_app = config.get("app", dict())
        self.APPLICATION_NAME = config_app.get("name", "sample")
        self.SERVICE_ID = uuid.uuid4().hex
        self.SERVICE_INSECURE_HOST = config_app.get("host", self.__SERVICE_DEFAULT_HOST)
        self.SERVICE_INSECURE_PORT = config_app.get("port", self.__SERVICE_DEFAULT_PORT)
        self.SERVICE_LEVEL = config_app.get("service_level", "development")
        config_model = config.get("model", dict())
        model_mode = config_model.get("mode", "local")
        self.MODEL_MODE_ENUM = ModelModeEnum.to_Enum(model_mode)
        if self.MODEL_MODE_ENUM == ModelModeEnum.LOCAL:
            config_model_mode = config_model.get(model_mode, dict())
            self.MODEL_FILE_PATH = config_model_mode.get("filepath", "model/default.model")
        elif self.MODEL_MODE_ENUM == ModelModeEnum.CEPH_S3:
            config_model_mode = config_model.get(model_mode, dict())
            self.MODEL_FILE_PATH = config_model_mode.get("filepath", "model/default.model")
            self.CEPH_ACCESS_KEY = config_model_mode.get("access_key")
            self.CEPH_SECRET_KEY = config_model_mode.get("secret_key")
            self.CEPH_HOST = config_model_mode.get("host")
            self.CEPH_PORT = config_model_mode.get("port", self.__CEPH_DEFAULT_PORT)
            self.CEPH_IS_SECURE = config_model_mode.get("is_secure", False)
            self.CEPH_BUCKET_NAME = config_model_mode.get("bucket")
        elif self.MODEL_MODE_ENUM == ModelModeEnum.AWS_S3:
            config_model_mode = config_model.get(model_mode, dict())
            self.MODEL_FILE_PATH = config_model_mode.get("filepath", "model/default.model")
            self.AWS_ACCESS_KEY = config_model_mode.get("access_key")
            self.AWS_SECRET_KEY = config_model_mode.get("secret_key")
            self.AWS_BUCKET_NAME = config_model_mode.get("bucket")
        elif self.MODEL_MODE_ENUM == ModelModeEnum.GCS:
            config_model_mode = config_model.get(model_mode, dict())
            self.MODEL_FILE_PATH = config_model_mode.get("filepath", "model/default.model")
            self.GCS_ACCESS_KEY = config_model_mode.get("access_key")
            self.GCS_SECRET_KEY = config_model_mode.get("secret_key")
            self.GCS_BUCKET_NAME = config_model_mode.get("bucket")
        else:
            raise ValueError("'{}' is not supported as ModelModeEnum".format(model_mode))

    def __load_from_env(self):
        self.DEBUG_MODE = os.getenv("REKCURD_DEBUG_MODE", "True").lower() == 'true'
        self.APPLICATION_NAME = os.getenv("REKCURD_APPLICATION_NAME")
        self.SERVICE_ID = os.getenv("REKCURD_SERVICE_ID")
        self.SERVICE_INSECURE_HOST = os.getenv("REKCURD_SERVICE_INSECURE_HOST", self.__SERVICE_DEFAULT_HOST)
        self.SERVICE_INSECURE_PORT = int(os.getenv("REKCURD_SERVICE_INSECURE_PORT", self.__SERVICE_DEFAULT_PORT))
        self.SERVICE_LEVEL = os.getenv("REKCURD_SERVICE_LEVEL")
        model_mode = os.getenv("REKCURD_MODEL_MODE")
        self.MODEL_MODE_ENUM = ModelModeEnum.to_Enum(model_mode)
        self.MODEL_FILE_PATH = os.getenv("REKCURD_MODEL_FILE_PATH")
        self.CEPH_ACCESS_KEY = os.getenv("REKCURD_CEPH_ACCESS_KEY")
        self.CEPH_SECRET_KEY = os.getenv("REKCURD_CEPH_SECRET_KEY")
        self.CEPH_HOST = os.getenv("REKCURD_CEPH_HOST")
        self.CEPH_PORT = int(os.getenv("REKCURD_CEPH_PORT", str(self.__CEPH_DEFAULT_PORT)))
        self.CEPH_IS_SECURE = os.getenv("REKCURD_CEPH_IS_SECURE", "False").lower() == 'true'
        self.CEPH_BUCKET_NAME = os.getenv("REKCURD_CEPH_BUCKET_NAME")
        self.AWS_ACCESS_KEY = os.getenv("REKCURD_AWS_ACCESS_KEY")
        self.AWS_SECRET_KEY = os.getenv("REKCURD_AWS_SECRET_KEY")
        self.AWS_BUCKET_NAME = os.getenv("REKCURD_AWS_BUCKET_NAME")
        self.GCS_ACCESS_KEY = os.getenv("REKCURD_GCS_ACCESS_KEY")
        self.GCS_SECRET_KEY = os.getenv("REKCURD_GCS_SECRET_KEY")
        self.GCS_BUCKET_NAME = os.getenv("REKCURD_GCS_BUCKET_NAME")



# -*- coding: utf-8 -*-


import os
import rekcurd
import shutil
import stat

from pathlib import Path

from .errors import CommandError


def _make_writeable(filename):
    """
    Make sure that the file is writeable.
    Useful if our source is read-only.
    """
    if not os.access(filename, os.W_OK):
        st = os.stat(filename)
        new_permissions = stat.S_IMODE(st.st_mode) | stat.S_IWUSR
        os.chmod(filename, new_permissions)


def startapp_handler(args):
    if Path(args.name).name != args.name:
        raise TypeError("Invalid project name: "+args.name)
    destination = Path(args.dir, args.name)
    if destination.exists():
        raise CommandError("'{}' already exists".format(destination.absolute()))
    destination.mkdir(parents=True, exist_ok=False)

    base_name = "RekcurdAppTemplate"
    template_suffix = "-tpl"
    template_dir = Path(rekcurd.__path__[0], 'template')

    for root, dirs, files in os.walk(template_dir):
        for dirname in dirs[:]:
            if dirname.startswith('.') or dirname == '__pycache__':
                dirs.remove(dirname)

        for filename in files:
            if filename.endswith(('.pyo', '.pyc', '.py.class')):
                # Ignore some files as they cause various breakages.
                continue
            old_path = Path(root, filename)
            new_path = Path(destination, filename)
            if str(new_path).endswith(template_suffix):
                new_path = Path(str(new_path)[:-len(template_suffix)])

            if new_path.exists():
                raise CommandError("{} already exists, overlaying a "
                                   "project into an existing directory"
                                   "won't replace conflicting files".format(new_path))

            with old_path.open(mode='r', encoding='utf-8') as template_file:
                content = template_file.read().replace(base_name, args.name)
            with new_path.open(mode='w', encoding='utf-8') as new_file:
                new_file.write(content)

            try:
                shutil.copymode(str(old_path), str(new_path))
                _make_writeable(str(new_path))
            except OSError:
                print(
                    "Notice: Couldn't set permission bits on {}. You're "
                    "probably using an uncommon filesystem setup. No "
                    "problem.".format(new_path)
                )



# coding: utf-8


import json
from typing import Union, List, Dict, NamedTuple

from .rekcurd_config import RekcurdConfig, ModelModeEnum


PredictInput = Union[str, bytes, List[str], List[int], List[float]]
PredictLabel = Union[str, bytes, List[str], List[int], List[float]]
PredictScore = Union[float, List[float]]


class PredictResult:
    def __init__(self, label: PredictLabel, score: PredictScore, option: dict = None):
        self.label = label
        self.score = score
        self.option = json.dumps(option) if option is not None else '{}'


class EvaluateResult:
    def __init__(self, num: int, accuracy: float, precision: List[float],
                 recall: List[float], fvalue: List[float], label: List[PredictLabel],
                 option: Dict[str, float] = {}):
        self.num = num
        self.accuracy = accuracy
        self.precision = precision
        self.recall = recall
        self.fvalue = fvalue
        self.label = label
        self.option = option


class EvaluateResultDetail(NamedTuple):
    result: PredictResult
    is_correct: bool


class EvaluateDetail(NamedTuple):
    input: PredictInput
    label: PredictLabel
    result: EvaluateResultDetail


incoming_headers = [
    'x-request-id', 'x-b3-traceid', 'x-b3-spanid', 'x-b3-parentspanid',
    'x-b3-sampled', 'x-b3-flags', 'x-ot-span-context']

def getForwardHeaders(incoming: list) -> list:
    headers = list()
    for k,v in incoming:
        if k in incoming_headers:
            headers.append((k, v))
    return headers



# coding: utf-8


from abc import ABCMeta, abstractmethod
from pathlib import Path

from rekcurd.utils import RekcurdConfig


def convert_to_valid_path(filepath: str) -> Path:
    """
    Remove a root path prefix "/", and a relative path "." and "..".
    :param filepath:
    :return:
    """
    valid_factors = [factor for factor in filepath.split("/") if factor and factor != ".."]
    return Path(*valid_factors)


class DataHandler(metaclass=ABCMeta):
    """Interface class.
    """
    LOCAL_MODEL_DIR: str = None
    MODEL_FILE_NAME: str = None
    LOCAL_EVAL_DIR: str = None

    def __init__(self, config: RekcurdConfig):
        valid_model_path = convert_to_valid_path(config.MODEL_FILE_PATH)
        self.LOCAL_MODEL_DIR = str(Path("rekcurd-model", valid_model_path.parent))
        self.MODEL_FILE_NAME = valid_model_path.name
        self.LOCAL_EVAL_DIR = "rekcurd-eval"

    @abstractmethod
    def download(self, remote_filepath: str, local_filepath: str) -> None:
        raise NotImplemented()

    @abstractmethod
    def upload(self, remote_filepath: str, local_filepath: str) -> None:
        raise NotImplemented()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import json
from abc import ABCMeta, abstractmethod


class SystemLoggerInterface(metaclass=ABCMeta):
    @abstractmethod
    def exception(self, message: str) -> None:
        raise NotImplemented()

    @abstractmethod
    def error(self, message: str) -> None:
        raise NotImplemented()

    @abstractmethod
    def debug(self, message: str) -> None:
        raise NotImplemented()

    @abstractmethod
    def info(self, message: str) -> None:
        raise NotImplemented()

    @abstractmethod
    def warn(self, message: str) -> None:
        raise NotImplemented()


class ServiceLoggerInterface(metaclass=ABCMeta):
    @abstractmethod
    def emit(self, request, response, suppress_log_inout: bool = False) -> None:
        raise NotImplemented()

    # noinspection PyMethodMayBeStatic
    def to_str_from_request(self, request) -> str:
        tmp = {'option': request.option.val}
        if isinstance(request.input, (str, bytes)):
            tmp['input'] = str(request.input)
        else:
            tmp['input'] = list(request.input)
        return json.dumps(tmp)

    # noinspection PyMethodMayBeStatic
    def to_str_from_response(self, response) -> str:
        tmp = {'option': response.option.val}
        if isinstance(response.output, (str, bytes)):
            tmp['output'] = str(response.output)
            tmp['score'] = response.score
        else:
            tmp['output'] = list(response.output)
            tmp['score'] = list(response.score)
        return json.dumps(tmp)



# coding: utf-8


from .rekcurd_worker import Rekcurd, RekcurdPack
from .rekcurd_worker_servicer import RekcurdInput, RekcurdOutput, RekcurdWorkerServicer
from .rekcurd_dashboard_servicer import RekcurdDashboardServicer



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
import sys
import time
from socket import gethostname

from fluent import handler, sender

from rekcurd.utils import RekcurdConfig
from .logger_interface import SystemLoggerInterface, ServiceLoggerInterface


class FluentSystemLogger(SystemLoggerInterface):

    def __init__(self,
                 logger_name: str = 'rekcurd.system',
                 log_level: int = None,
                 config: RekcurdConfig = RekcurdConfig()) -> None:
        """
        Constructor
        :param logger_name:
        :param log_level:
        :param config: RekcurdConfig
        """
        self.config = config
        log_level = int(log_level or logging.DEBUG if config.DEBUG_MODE else logging.NOTSET)
        application_name = config.APPLICATION_NAME
        service_level = config.SERVICE_LEVEL
        self.log = logging.getLogger(logger_name)
        self.log.setLevel(log_level)
        self.log.handlers = []
        self.log.addHandler(self.__init_fluent_handler(application_name, service_level, log_level))

    def __init_fluent_handler(self, application_name: str, service_level: str, log_level: int):
        custom_format = {
            'host': gethostname(),
            'short_message': '%(message)s',
            'timestamp': '%(created)d.%(msecs)d',
            'level': '%(loglevel)d',
            'service': 'rekcurd',
            'ml_service': application_name,
            'service_level': service_level
        }
        fluent_handler = handler.FluentHandler('rekcurd')
        formatter = handler.FluentRecordFormatter(custom_format)
        fluent_handler.setFormatter(formatter)
        fluent_handler.setLevel(log_level)
        return fluent_handler

    def exception(self, message: str) -> None:
        """
        emits exception to log
        :param message: error message
        """
        self.log.error(message, exc_info=sys.exc_info(), stack_info=True, extra={'loglevel': 3})

    def error(self, message: str) -> None:
        """
        emits error log
        :param message: log
        """
        self.log.error(message, extra={'loglevel': 3})

    def debug(self, message: str) -> None:
        """
        emits debug log
        :param message: log
        """
        self.log.debug(message, extra={'loglevel': 7})

    def info(self, message: str) -> None:
        """
        emits info log
        :param message: log
        """
        self.log.info(message, extra={'loglevel': 6})

    def warn(self, message: str) -> None:
        """
        emits warn log
        :param message: log
        """
        self.log.warning(message, extra={'loglevel': 4})


class FluentServiceLogger(ServiceLoggerInterface):

    def __init__(self,
                 logger_name: str = 'rekcurd.service',
                 log_level: int = None,
                 config: RekcurdConfig = RekcurdConfig()):
        """
        Constructor
        :param logger_name:
        :param log_level:
        :param config: RekcurdConfig
        """
        self.logger_name = logger_name
        self.log_level = int(log_level or logging.DEBUG)
        self.config = config
        self.ml_service = config.APPLICATION_NAME
        self.service_level = config.SERVICE_LEVEL
        self.logger = sender.FluentSender(logger_name)

    def emit(self, request, response, suppress_log_inout: bool = False) -> None:
        """
        emits service log
        :param request:
        :param response:
        :param suppress_log_inout:
        :return:
        """
        try:
            if suppress_log_inout:
                ml_input = ''
                ml_output = ''
            else:
                ml_input = super().to_str_from_request(request)
                ml_output = super().to_str_from_response(response)

            self.logger.emit(None, {
                'host': gethostname(),
                'short_message': 'prediction result.',
                'timestamp': int(time.time() * 1000) / 1000,
                'level': self.log_level,
                'service': 'rekcurd',
                'ml_service': self.ml_service,
                'service_level': self.service_level,
                'ml_input': ml_input,
                'ml_output': ml_output
            })
        except:
            try:
                FluentSystemLogger(self.logger_name, self.log_level, self.config).exception("can't write log")
            except:
                pass



# coding: utf-8


import boto3

from .data_handler import DataHandler
from rekcurd.utils import RekcurdConfig


class AwsS3Handler(DataHandler):
    """AwsS3Handler
    """
    def __init__(self, config: RekcurdConfig):
        super(AwsS3Handler, self).__init__(config)
        self._resource = boto3.resource(
            's3',
            aws_access_key_id=config.AWS_ACCESS_KEY,
            aws_secret_access_key=config.AWS_SECRET_KEY,
        )
        self._bucket = config.AWS_BUCKET_NAME

    def download(self, remote_filepath: str, local_filepath: str) -> None:
        self._resource.Bucket(self._bucket).download_file(remote_filepath, local_filepath)

    def upload(self, remote_filepath: str, local_filepath: str) -> None:
        self._resource.Bucket(self._bucket).upload_file(local_filepath, remote_filepath)



#!/usr/bin/python
# -*- coding: utf-8 -*-


import traceback

import grpc
import types
import pickle
import sys

from grpc import ServicerContext
from typing import Iterator, Union, List

from .rekcurd_worker import RekcurdPack
from rekcurd.protobuf import rekcurd_pb2, rekcurd_pb2_grpc
from rekcurd.utils import PredictInput, PredictLabel, PredictScore


def error_handling(error_response):
    """ Decorator for handling error

    Apply following processing on Servicer methods
    to handle errors.

    - DB transaction decorating for Servicer class.
      Confirm to call :func:``db.session.commit``
      on success operation and :func:``db.session.rollback``
      on fail.
    - Error setup for gRPC errors
    - Call :func:``on_error`` method (if defined) in the class
      to postprocess something on error

    Parameters
    ----------
    error_response
        gRPC response instance on error

    """

    def _wrapper_maker(func):
        def _wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as error:
                # gRPC
                context = args[2]
                context.set_code(grpc.StatusCode.UNKNOWN)
                context.set_details(str(error))

                servicer = args[0]
                if hasattr(servicer, 'on_error'):
                    assert isinstance(servicer.on_error, types.MethodType), \
                        'You must define on_error as method'
                    servicer.on_error(error)
                return error_response

        return _wrapper

    return _wrapper_maker


class RekcurdDashboardServicer(rekcurd_pb2_grpc.RekcurdDashboardServicer):
    """ gRPC servicer to manage environment

    - Applications
        Machine leagning applications
    - Services
        Unit to deploy machine learning models
        (Corresponding to Service of K8S)
    - Models
        Machine learning model
    """

    CHUNK_SIZE = 100
    BYTE_LIMIT = 4190000

    def __init__(self, rekcurd_pack: RekcurdPack):
        self.rekcurd_pack = rekcurd_pack
        self.logger = rekcurd_pack.app.system_logger

    def on_error(self, error: Exception):
        """ Postprocessing on error

        For detail, see :func:``on_error``

        Parameters
        ----------
        error : Exception
            Error to be handled
        """
        self.logger.error(str(error))
        self.logger.error(traceback.format_exc())

    def ServiceInfo(self,
                    request: rekcurd_pb2.ServiceInfoRequest,
                    context: ServicerContext
                    ) -> rekcurd_pb2.ServiceInfoResponse:
        """ Get service info.

        :param request:
        :param context:
        :return:
        """
        self.logger.info("Run ServiceInfo.")
        return rekcurd_pb2.ServiceInfoResponse(
            application_name=self.rekcurd_pack.app.config.APPLICATION_NAME,
            service_name=self.rekcurd_pack.app.config.SERVICE_ID,
            service_level=self.rekcurd_pack.app.config.SERVICE_LEVEL)

    @error_handling(rekcurd_pb2.ModelResponse(status=0, message='Error: Uploading model file.'))
    def UploadModel(self,
                    request_iterator: Iterator[rekcurd_pb2.UploadModelRequest],
                    context: ServicerContext
                    ) -> rekcurd_pb2.ModelResponse:
        """ Upload your latest ML model.

        :param request_iterator:
        :param context:
        :return:
        """
        self.logger.info("Run UploadModel.")
        self.rekcurd_pack.app.data_server.upload_model(request_iterator)
        return rekcurd_pb2.ModelResponse(status=1,
                                         message='Success: Uploading model file.')

    @error_handling(rekcurd_pb2.ModelResponse(status=0, message='Error: Switching model file.'))
    def SwitchModel(self,
                    request: rekcurd_pb2.SwitchModelRequest,
                    context: ServicerContext
                    ) -> rekcurd_pb2.ModelResponse:
        """ Switch your ML model.

        :param request:
        :param context:
        :return:
        """
        self.logger.info("Run SwitchModel.")
        filepath = request.path
        local_filepath = self.rekcurd_pack.app.data_server.switch_model(filepath)
        self.rekcurd_pack.predictor = self.rekcurd_pack.app.load_model(local_filepath)
        return rekcurd_pb2.ModelResponse(status=1,
                                         message='Success: Switching model file.')

    @error_handling(rekcurd_pb2.EvaluateModelResponse())
    def EvaluateModel(self,
                      request_iterator: Iterator[rekcurd_pb2.EvaluateModelRequest],
                      context: ServicerContext
                      ) -> rekcurd_pb2.EvaluateModelResponse:
        """ Evaluate your ML model and save result.

        :param request_iterator:
        :param context:
        :return:
        """
        self.logger.info("Run EvaluateModel.")
        first_req = next(request_iterator)
        data_path = first_req.data_path
        result_path = first_req.result_path

        local_data_path = self.rekcurd_pack.app.data_server.get_evaluation_data_path(data_path)
        evaluate_result_gen = self.rekcurd_pack.app.evaluate(self.rekcurd_pack.predictor, local_data_path)
        result = self.rekcurd_pack.app.data_server.upload_evaluation_result(evaluate_result_gen, result_path)
        label_ios = [self.get_io_by_type(l) for l in result.label]
        metrics = rekcurd_pb2.EvaluationMetrics(num=result.num,
                                                accuracy=result.accuracy,
                                                precision=result.precision,
                                                recall=result.recall,
                                                fvalue=result.fvalue,
                                                option=result.option,
                                                label=label_ios)
        return rekcurd_pb2.EvaluateModelResponse(metrics=metrics)

    @error_handling(rekcurd_pb2.UploadEvaluationDataResponse(status=0, message='Error: Uploading evaluation data.'))
    def UploadEvaluationData(self,
                             request_iterator: Iterator[rekcurd_pb2.UploadEvaluationDataRequest],
                             context: ServicerContext
                             ) -> rekcurd_pb2.UploadEvaluationDataResponse:
        """ Save evaluation data

        :param request_iterator:
        :param context:
        :return:
        """
        self.logger.info("Run UploadEvaluationData.")
        self.rekcurd_pack.app.data_server.upload_evaluation_data(request_iterator)
        return rekcurd_pb2.UploadEvaluationDataResponse(status=1,
                                                        message='Success: Uploading evaluation data.')

    @error_handling(rekcurd_pb2.EvaluationResultResponse())
    def EvaluationResult(self,
                         request: rekcurd_pb2.EvaluationResultRequest,
                         context: ServicerContext
                         ) -> Iterator[rekcurd_pb2.EvaluationResultResponse]:
        """ Return saved evaluation result
        """
        self.logger.info("Run EvaluationResult.")
        data_path = request.data_path
        result_path = request.result_path
        local_data_path = self.rekcurd_pack.app.data_server.get_evaluation_data_path(data_path)
        local_result_detail_path = self.rekcurd_pack.app.data_server.get_eval_result_detail(result_path)

        # TODO: deprecated. remove metrics from response in the next gRPC spec
        metrics = rekcurd_pb2.EvaluationMetrics()

        detail_chunks = []
        detail_chunk = []
        metrics_size = sys.getsizeof(metrics)
        with open(local_result_detail_path, 'rb') as detail_file:
            def generate_result_detail():
                try:
                    while True:
                        yield pickle.load(detail_file)
                except EOFError:
                    pass
            for detail in self.rekcurd_pack.app.get_evaluate_detail(local_data_path, generate_result_detail()):
                detail_chunk.append(rekcurd_pb2.EvaluationResultResponse.Detail(
                    input=self.get_io_by_type(detail.input),
                    label=self.get_io_by_type(detail.label),
                    output=self.get_io_by_type(detail.result.result.label),
                    score=self.get_score_by_type(detail.result.result.score),
                    is_correct=detail.result.is_correct
                ))
                if len(detail_chunk) == self.CHUNK_SIZE:
                    if metrics_size + sys.getsizeof(detail_chunk + detail_chunks) < self.BYTE_LIMIT:
                        detail_chunks.extend(detail_chunk)
                    else:
                        yield rekcurd_pb2.EvaluationResultResponse(metrics=metrics, detail=detail_chunks)
                        detail_chunks = detail_chunk
                    detail_chunk = []

        if len(detail_chunks + detail_chunk) > 0:
            if metrics_size + sys.getsizeof(detail_chunk + detail_chunks) < self.BYTE_LIMIT:
                detail_chunks.extend(detail_chunk)
                yield rekcurd_pb2.EvaluationResultResponse(metrics=metrics, detail=detail_chunks)
            else:
                yield rekcurd_pb2.EvaluationResultResponse(metrics=metrics, detail=detail_chunks)
                yield rekcurd_pb2.EvaluationResultResponse(metrics=metrics, detail=detail_chunk)

    def get_io_by_type(self, io: Union[PredictInput, PredictLabel]) -> rekcurd_pb2.IO:
        if not isinstance(io, list):
            io = [io]

        if len(io) == 0:
            return rekcurd_pb2.IO(tensor=rekcurd_pb2.Tensor())
        if isinstance(io[0], str):
            return rekcurd_pb2.IO(str=rekcurd_pb2.ArrString(val=io))
        else:
            return rekcurd_pb2.IO(tensor=rekcurd_pb2.Tensor(shape=[1], val=io))

    def get_score_by_type(self, score: PredictScore) -> List[float]:
        if isinstance(score, list):
            return score
        else:
            return [score]



# -*- coding: utf-8 -*-


import argparse

from .startapp_handler import startapp_handler
from rekcurd import _version


def create_parser():
    parser = argparse.ArgumentParser(description='rekcurd command')
    parser.add_argument(
        '--version', '-v', action='version', version=_version.__version__)
    subparsers = parser.add_subparsers()

    # startapp
    parser_startapp = subparsers.add_parser(
        'startapp', help='see `rekcurd startapp -h`')
    parser_startapp.add_argument(
        'name', help='Name of the application or project.')
    parser_startapp.add_argument(
        '--dir', required=False, help='Optional destination directory', default='./')
    parser_startapp.set_defaults(handler=startapp_handler)

    return parser


def main() -> None:
    parser = create_parser()
    args = parser.parse_args()

    if hasattr(args, 'handler'):
        args.handler(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



# coding: utf-8


import boto
import boto.s3.connection

from .data_handler import DataHandler
from rekcurd.utils import RekcurdConfig


class CephHandler(DataHandler):
    """CephHandler
    """
    def __init__(self, config: RekcurdConfig):
        super(CephHandler, self).__init__(config)
        self._conn = boto.connect_s3(
            aws_access_key_id=config.CEPH_ACCESS_KEY,
            aws_secret_access_key=config.CEPH_SECRET_KEY,
            host=config.CEPH_HOST,
            port=config.CEPH_PORT,
            is_secure=config.CEPH_IS_SECURE,
            calling_format=boto.s3.connection.OrdinaryCallingFormat())
        self._bucket = config.CEPH_BUCKET_NAME

    def download(self, remote_filepath: str, local_filepath: str) -> None:
        bucket = self._conn.get_bucket(self._bucket)
        key = bucket.get_key(remote_filepath)
        key.get_contents_to_filename(local_filepath)

    def upload(self, remote_filepath: str, local_filepath: str) -> None:
        bucket = self._conn.get_bucket(self._bucket)
        key = bucket.new_key(remote_filepath)
        key.set_contents_from_filename(local_filepath, replace=False)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import rekcurd_pb2 as rekcurd__pb2


class HealthStub(object):
  """gRPC official health check service.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.Check = channel.unary_unary(
        '/rekcurd.Health/Check',
        request_serializer=rekcurd__pb2.HealthCheckRequest.SerializeToString,
        response_deserializer=rekcurd__pb2.HealthCheckResponse.FromString,
        )


class HealthServicer(object):
  """gRPC official health check service.
  """

  def Check(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_HealthServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'Check': grpc.unary_unary_rpc_method_handler(
          servicer.Check,
          request_deserializer=rekcurd__pb2.HealthCheckRequest.FromString,
          response_serializer=rekcurd__pb2.HealthCheckResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'rekcurd.Health', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class RekcurdDashboardStub(object):
  """Rekcurd Dashboard service.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ServiceInfo = channel.unary_unary(
        '/rekcurd.RekcurdDashboard/ServiceInfo',
        request_serializer=rekcurd__pb2.ServiceInfoRequest.SerializeToString,
        response_deserializer=rekcurd__pb2.ServiceInfoResponse.FromString,
        )
    self.UploadModel = channel.stream_unary(
        '/rekcurd.RekcurdDashboard/UploadModel',
        request_serializer=rekcurd__pb2.UploadModelRequest.SerializeToString,
        response_deserializer=rekcurd__pb2.ModelResponse.FromString,
        )
    self.SwitchModel = channel.unary_unary(
        '/rekcurd.RekcurdDashboard/SwitchModel',
        request_serializer=rekcurd__pb2.SwitchModelRequest.SerializeToString,
        response_deserializer=rekcurd__pb2.ModelResponse.FromString,
        )
    self.UploadEvaluationData = channel.stream_unary(
        '/rekcurd.RekcurdDashboard/UploadEvaluationData',
        request_serializer=rekcurd__pb2.UploadEvaluationDataRequest.SerializeToString,
        response_deserializer=rekcurd__pb2.UploadEvaluationDataResponse.FromString,
        )
    self.EvaluateModel = channel.stream_unary(
        '/rekcurd.RekcurdDashboard/EvaluateModel',
        request_serializer=rekcurd__pb2.EvaluateModelRequest.SerializeToString,
        response_deserializer=rekcurd__pb2.EvaluateModelResponse.FromString,
        )
    self.EvaluationResult = channel.unary_stream(
        '/rekcurd.RekcurdDashboard/EvaluationResult',
        request_serializer=rekcurd__pb2.EvaluationResultRequest.SerializeToString,
        response_deserializer=rekcurd__pb2.EvaluationResultResponse.FromString,
        )


class RekcurdDashboardServicer(object):
  """Rekcurd Dashboard service.
  """

  def ServiceInfo(self, request, context):
    """Get Service Info.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UploadModel(self, request_iterator, context):
    """Upload your latest ML model.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SwitchModel(self, request, context):
    """Switch your ML model to run.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UploadEvaluationData(self, request_iterator, context):
    """Upload data for evaluation
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def EvaluateModel(self, request_iterator, context):
    """Evaluate the accuracy of your ML model.
    Deprecated: `stream` is deprecated. upload evaluation data by using UploadEvaluationData
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def EvaluationResult(self, request, context):
    """Get detailed result of EvaluateModel
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_RekcurdDashboardServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ServiceInfo': grpc.unary_unary_rpc_method_handler(
          servicer.ServiceInfo,
          request_deserializer=rekcurd__pb2.ServiceInfoRequest.FromString,
          response_serializer=rekcurd__pb2.ServiceInfoResponse.SerializeToString,
      ),
      'UploadModel': grpc.stream_unary_rpc_method_handler(
          servicer.UploadModel,
          request_deserializer=rekcurd__pb2.UploadModelRequest.FromString,
          response_serializer=rekcurd__pb2.ModelResponse.SerializeToString,
      ),
      'SwitchModel': grpc.unary_unary_rpc_method_handler(
          servicer.SwitchModel,
          request_deserializer=rekcurd__pb2.SwitchModelRequest.FromString,
          response_serializer=rekcurd__pb2.ModelResponse.SerializeToString,
      ),
      'UploadEvaluationData': grpc.stream_unary_rpc_method_handler(
          servicer.UploadEvaluationData,
          request_deserializer=rekcurd__pb2.UploadEvaluationDataRequest.FromString,
          response_serializer=rekcurd__pb2.UploadEvaluationDataResponse.SerializeToString,
      ),
      'EvaluateModel': grpc.stream_unary_rpc_method_handler(
          servicer.EvaluateModel,
          request_deserializer=rekcurd__pb2.EvaluateModelRequest.FromString,
          response_serializer=rekcurd__pb2.EvaluateModelResponse.SerializeToString,
      ),
      'EvaluationResult': grpc.unary_stream_rpc_method_handler(
          servicer.EvaluationResult,
          request_deserializer=rekcurd__pb2.EvaluationResultRequest.FromString,
          response_serializer=rekcurd__pb2.EvaluationResultResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'rekcurd.RekcurdDashboard', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class RekcurdWorkerStub(object):
  """Rekcurd worker service.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.Predict_String_String = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_String_String',
        request_serializer=rekcurd__pb2.StringInput.SerializeToString,
        response_deserializer=rekcurd__pb2.StringOutput.FromString,
        )
    self.Predict_String_Bytes = channel.unary_stream(
        '/rekcurd.RekcurdWorker/Predict_String_Bytes',
        request_serializer=rekcurd__pb2.StringInput.SerializeToString,
        response_deserializer=rekcurd__pb2.BytesOutput.FromString,
        )
    self.Predict_String_ArrInt = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_String_ArrInt',
        request_serializer=rekcurd__pb2.StringInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrIntOutput.FromString,
        )
    self.Predict_String_ArrFloat = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_String_ArrFloat',
        request_serializer=rekcurd__pb2.StringInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrFloatOutput.FromString,
        )
    self.Predict_String_ArrString = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_String_ArrString',
        request_serializer=rekcurd__pb2.StringInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrStringOutput.FromString,
        )
    self.Predict_Bytes_String = channel.stream_unary(
        '/rekcurd.RekcurdWorker/Predict_Bytes_String',
        request_serializer=rekcurd__pb2.BytesInput.SerializeToString,
        response_deserializer=rekcurd__pb2.StringOutput.FromString,
        )
    self.Predict_Bytes_Bytes = channel.stream_stream(
        '/rekcurd.RekcurdWorker/Predict_Bytes_Bytes',
        request_serializer=rekcurd__pb2.BytesInput.SerializeToString,
        response_deserializer=rekcurd__pb2.BytesOutput.FromString,
        )
    self.Predict_Bytes_ArrInt = channel.stream_unary(
        '/rekcurd.RekcurdWorker/Predict_Bytes_ArrInt',
        request_serializer=rekcurd__pb2.BytesInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrIntOutput.FromString,
        )
    self.Predict_Bytes_ArrFloat = channel.stream_unary(
        '/rekcurd.RekcurdWorker/Predict_Bytes_ArrFloat',
        request_serializer=rekcurd__pb2.BytesInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrFloatOutput.FromString,
        )
    self.Predict_Bytes_ArrString = channel.stream_unary(
        '/rekcurd.RekcurdWorker/Predict_Bytes_ArrString',
        request_serializer=rekcurd__pb2.BytesInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrStringOutput.FromString,
        )
    self.Predict_ArrInt_String = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrInt_String',
        request_serializer=rekcurd__pb2.ArrIntInput.SerializeToString,
        response_deserializer=rekcurd__pb2.StringOutput.FromString,
        )
    self.Predict_ArrInt_Bytes = channel.unary_stream(
        '/rekcurd.RekcurdWorker/Predict_ArrInt_Bytes',
        request_serializer=rekcurd__pb2.ArrIntInput.SerializeToString,
        response_deserializer=rekcurd__pb2.BytesOutput.FromString,
        )
    self.Predict_ArrInt_ArrInt = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrInt_ArrInt',
        request_serializer=rekcurd__pb2.ArrIntInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrIntOutput.FromString,
        )
    self.Predict_ArrInt_ArrFloat = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrInt_ArrFloat',
        request_serializer=rekcurd__pb2.ArrIntInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrFloatOutput.FromString,
        )
    self.Predict_ArrInt_ArrString = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrInt_ArrString',
        request_serializer=rekcurd__pb2.ArrIntInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrStringOutput.FromString,
        )
    self.Predict_ArrFloat_String = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrFloat_String',
        request_serializer=rekcurd__pb2.ArrFloatInput.SerializeToString,
        response_deserializer=rekcurd__pb2.StringOutput.FromString,
        )
    self.Predict_ArrFloat_Bytes = channel.unary_stream(
        '/rekcurd.RekcurdWorker/Predict_ArrFloat_Bytes',
        request_serializer=rekcurd__pb2.ArrFloatInput.SerializeToString,
        response_deserializer=rekcurd__pb2.BytesOutput.FromString,
        )
    self.Predict_ArrFloat_ArrInt = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrFloat_ArrInt',
        request_serializer=rekcurd__pb2.ArrFloatInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrIntOutput.FromString,
        )
    self.Predict_ArrFloat_ArrFloat = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrFloat_ArrFloat',
        request_serializer=rekcurd__pb2.ArrFloatInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrFloatOutput.FromString,
        )
    self.Predict_ArrFloat_ArrString = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrFloat_ArrString',
        request_serializer=rekcurd__pb2.ArrFloatInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrStringOutput.FromString,
        )
    self.Predict_ArrString_String = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrString_String',
        request_serializer=rekcurd__pb2.ArrStringInput.SerializeToString,
        response_deserializer=rekcurd__pb2.StringOutput.FromString,
        )
    self.Predict_ArrString_Bytes = channel.unary_stream(
        '/rekcurd.RekcurdWorker/Predict_ArrString_Bytes',
        request_serializer=rekcurd__pb2.ArrStringInput.SerializeToString,
        response_deserializer=rekcurd__pb2.BytesOutput.FromString,
        )
    self.Predict_ArrString_ArrInt = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrString_ArrInt',
        request_serializer=rekcurd__pb2.ArrStringInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrIntOutput.FromString,
        )
    self.Predict_ArrString_ArrFloat = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrString_ArrFloat',
        request_serializer=rekcurd__pb2.ArrStringInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrFloatOutput.FromString,
        )
    self.Predict_ArrString_ArrString = channel.unary_unary(
        '/rekcurd.RekcurdWorker/Predict_ArrString_ArrString',
        request_serializer=rekcurd__pb2.ArrStringInput.SerializeToString,
        response_deserializer=rekcurd__pb2.ArrStringOutput.FromString,
        )


class RekcurdWorkerServicer(object):
  """Rekcurd worker service.
  """

  def Predict_String_String(self, request, context):
    """Input string, output string.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_String_Bytes(self, request, context):
    """Input string, output bytes.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_String_ArrInt(self, request, context):
    """Input string, output array[int].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_String_ArrFloat(self, request, context):
    """Input string, output array[float].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_String_ArrString(self, request, context):
    """Input string, output array[string].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_Bytes_String(self, request_iterator, context):
    """Input bytes, output string.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_Bytes_Bytes(self, request_iterator, context):
    """Input bytes, output bytes.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_Bytes_ArrInt(self, request_iterator, context):
    """Input bytes, output array[int].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_Bytes_ArrFloat(self, request_iterator, context):
    """Input bytes, output array[float].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_Bytes_ArrString(self, request_iterator, context):
    """Input bytes, output array[string].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrInt_String(self, request, context):
    """Input array[int], output string.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrInt_Bytes(self, request, context):
    """Input array[int], output bytes.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrInt_ArrInt(self, request, context):
    """Input array[int], output array[int].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrInt_ArrFloat(self, request, context):
    """Input array[int], output array[float].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrInt_ArrString(self, request, context):
    """Input array[int], output array[string].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrFloat_String(self, request, context):
    """Input array[float], output string.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrFloat_Bytes(self, request, context):
    """Input array[float], output bytes.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrFloat_ArrInt(self, request, context):
    """Input array[float], output array[int].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrFloat_ArrFloat(self, request, context):
    """Input array[float], output array[float].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrFloat_ArrString(self, request, context):
    """Input array[float], output array[string].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrString_String(self, request, context):
    """Input array[string], output string.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrString_Bytes(self, request, context):
    """Input array[string], output bytes.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrString_ArrInt(self, request, context):
    """Input array[string], output array[int].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrString_ArrFloat(self, request, context):
    """Input array[string], output array[float].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Predict_ArrString_ArrString(self, request, context):
    """Input array[string], output array[string].
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_RekcurdWorkerServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'Predict_String_String': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_String_String,
          request_deserializer=rekcurd__pb2.StringInput.FromString,
          response_serializer=rekcurd__pb2.StringOutput.SerializeToString,
      ),
      'Predict_String_Bytes': grpc.unary_stream_rpc_method_handler(
          servicer.Predict_String_Bytes,
          request_deserializer=rekcurd__pb2.StringInput.FromString,
          response_serializer=rekcurd__pb2.BytesOutput.SerializeToString,
      ),
      'Predict_String_ArrInt': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_String_ArrInt,
          request_deserializer=rekcurd__pb2.StringInput.FromString,
          response_serializer=rekcurd__pb2.ArrIntOutput.SerializeToString,
      ),
      'Predict_String_ArrFloat': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_String_ArrFloat,
          request_deserializer=rekcurd__pb2.StringInput.FromString,
          response_serializer=rekcurd__pb2.ArrFloatOutput.SerializeToString,
      ),
      'Predict_String_ArrString': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_String_ArrString,
          request_deserializer=rekcurd__pb2.StringInput.FromString,
          response_serializer=rekcurd__pb2.ArrStringOutput.SerializeToString,
      ),
      'Predict_Bytes_String': grpc.stream_unary_rpc_method_handler(
          servicer.Predict_Bytes_String,
          request_deserializer=rekcurd__pb2.BytesInput.FromString,
          response_serializer=rekcurd__pb2.StringOutput.SerializeToString,
      ),
      'Predict_Bytes_Bytes': grpc.stream_stream_rpc_method_handler(
          servicer.Predict_Bytes_Bytes,
          request_deserializer=rekcurd__pb2.BytesInput.FromString,
          response_serializer=rekcurd__pb2.BytesOutput.SerializeToString,
      ),
      'Predict_Bytes_ArrInt': grpc.stream_unary_rpc_method_handler(
          servicer.Predict_Bytes_ArrInt,
          request_deserializer=rekcurd__pb2.BytesInput.FromString,
          response_serializer=rekcurd__pb2.ArrIntOutput.SerializeToString,
      ),
      'Predict_Bytes_ArrFloat': grpc.stream_unary_rpc_method_handler(
          servicer.Predict_Bytes_ArrFloat,
          request_deserializer=rekcurd__pb2.BytesInput.FromString,
          response_serializer=rekcurd__pb2.ArrFloatOutput.SerializeToString,
      ),
      'Predict_Bytes_ArrString': grpc.stream_unary_rpc_method_handler(
          servicer.Predict_Bytes_ArrString,
          request_deserializer=rekcurd__pb2.BytesInput.FromString,
          response_serializer=rekcurd__pb2.ArrStringOutput.SerializeToString,
      ),
      'Predict_ArrInt_String': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrInt_String,
          request_deserializer=rekcurd__pb2.ArrIntInput.FromString,
          response_serializer=rekcurd__pb2.StringOutput.SerializeToString,
      ),
      'Predict_ArrInt_Bytes': grpc.unary_stream_rpc_method_handler(
          servicer.Predict_ArrInt_Bytes,
          request_deserializer=rekcurd__pb2.ArrIntInput.FromString,
          response_serializer=rekcurd__pb2.BytesOutput.SerializeToString,
      ),
      'Predict_ArrInt_ArrInt': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrInt_ArrInt,
          request_deserializer=rekcurd__pb2.ArrIntInput.FromString,
          response_serializer=rekcurd__pb2.ArrIntOutput.SerializeToString,
      ),
      'Predict_ArrInt_ArrFloat': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrInt_ArrFloat,
          request_deserializer=rekcurd__pb2.ArrIntInput.FromString,
          response_serializer=rekcurd__pb2.ArrFloatOutput.SerializeToString,
      ),
      'Predict_ArrInt_ArrString': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrInt_ArrString,
          request_deserializer=rekcurd__pb2.ArrIntInput.FromString,
          response_serializer=rekcurd__pb2.ArrStringOutput.SerializeToString,
      ),
      'Predict_ArrFloat_String': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrFloat_String,
          request_deserializer=rekcurd__pb2.ArrFloatInput.FromString,
          response_serializer=rekcurd__pb2.StringOutput.SerializeToString,
      ),
      'Predict_ArrFloat_Bytes': grpc.unary_stream_rpc_method_handler(
          servicer.Predict_ArrFloat_Bytes,
          request_deserializer=rekcurd__pb2.ArrFloatInput.FromString,
          response_serializer=rekcurd__pb2.BytesOutput.SerializeToString,
      ),
      'Predict_ArrFloat_ArrInt': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrFloat_ArrInt,
          request_deserializer=rekcurd__pb2.ArrFloatInput.FromString,
          response_serializer=rekcurd__pb2.ArrIntOutput.SerializeToString,
      ),
      'Predict_ArrFloat_ArrFloat': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrFloat_ArrFloat,
          request_deserializer=rekcurd__pb2.ArrFloatInput.FromString,
          response_serializer=rekcurd__pb2.ArrFloatOutput.SerializeToString,
      ),
      'Predict_ArrFloat_ArrString': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrFloat_ArrString,
          request_deserializer=rekcurd__pb2.ArrFloatInput.FromString,
          response_serializer=rekcurd__pb2.ArrStringOutput.SerializeToString,
      ),
      'Predict_ArrString_String': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrString_String,
          request_deserializer=rekcurd__pb2.ArrStringInput.FromString,
          response_serializer=rekcurd__pb2.StringOutput.SerializeToString,
      ),
      'Predict_ArrString_Bytes': grpc.unary_stream_rpc_method_handler(
          servicer.Predict_ArrString_Bytes,
          request_deserializer=rekcurd__pb2.ArrStringInput.FromString,
          response_serializer=rekcurd__pb2.BytesOutput.SerializeToString,
      ),
      'Predict_ArrString_ArrInt': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrString_ArrInt,
          request_deserializer=rekcurd__pb2.ArrStringInput.FromString,
          response_serializer=rekcurd__pb2.ArrIntOutput.SerializeToString,
      ),
      'Predict_ArrString_ArrFloat': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrString_ArrFloat,
          request_deserializer=rekcurd__pb2.ArrStringInput.FromString,
          response_serializer=rekcurd__pb2.ArrFloatOutput.SerializeToString,
      ),
      'Predict_ArrString_ArrString': grpc.unary_unary_rpc_method_handler(
          servicer.Predict_ArrString_ArrString,
          request_deserializer=rekcurd__pb2.ArrStringInput.FromString,
          response_serializer=rekcurd__pb2.ArrStringOutput.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'rekcurd.RekcurdWorker', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# -*- coding: utf-8 -*-


from .logger_interface import SystemLoggerInterface, ServiceLoggerInterface
from .logger_jsonlogger import JsonSystemLogger, JsonServiceLogger
from .logger_fluent import FluentSystemLogger, FluentServiceLogger



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
import sys
import time
from socket import gethostname

from pythonjsonlogger import jsonlogger

from rekcurd.utils import RekcurdConfig
from .logger_interface import SystemLoggerInterface, ServiceLoggerInterface


class JsonSystemLogger(SystemLoggerInterface):
    class JsonFormatter(jsonlogger.JsonFormatter):
        def parse(self):
            return [
                'host',
                'short_message',
                'timestamp',
                'level',
                'service',
                'ml_service',
                'service_level',
            ]

        def add_fields(self, log_record, record, message_dict):
            super().add_fields(log_record, record, message_dict)
            log_record['host'] = gethostname()
            log_record['timestamp'] = int(time.time() * 1000) / 1000
            log_record['service'] = 'rekcurd'

    def __init__(self,
                 logger_name: str = 'rekcurd.system',
                 log_level: int = None,
                 config: RekcurdConfig = RekcurdConfig()) -> None:
        """
        Constructor
        :param logger_name:
        :param log_level:
        :param config: RekcurdConfig
        """
        self.config = config
        log_level = int(log_level or logging.DEBUG if config.DEBUG_MODE else logging.NOTSET)
        self.ml_service = config.APPLICATION_NAME
        self.service_level = config.SERVICE_LEVEL
        self.log = logging.getLogger(logger_name)
        handler = logging.StreamHandler()
        formatter = self.JsonFormatter()
        handler.setFormatter(formatter)
        handler.setLevel(log_level)
        self.log.handlers = []
        self.log.addHandler(handler)
        self.log.setLevel(log_level)

    def exception(self, message: str) -> None:
        """
        emits exception to log
        :param message: error message
        """
        self.log.error(message, exc_info=sys.exc_info(), stack_info=True,
                       extra={'short_message': message, 'level': 3,
                              'ml_service': self.ml_service,
                              'service_level': self.service_level})

    def error(self, message: str) -> None:
        """
        emits error log
        :param message: log
        """
        self.log.error(message, extra={'short_message': message, 'level': 3,
                                       'ml_service': self.ml_service,
                                       'service_level': self.service_level})

    def debug(self, message: str) -> None:
        """
        emits debug log
        :param message: log
        """
        self.log.debug(message, extra={'short_message': message, 'level': 7,
                                       'ml_service': self.ml_service,
                                       'service_level': self.service_level})

    def info(self, message: str) -> None:
        """
        emits info log
        :param message: log
        """
        self.log.info(message, extra={'short_message': message, 'level': 6,
                                      'ml_service': self.ml_service,
                                      'service_level': self.service_level})

    def warn(self, message: str) -> None:
        """
        emits warn log
        :param message: log
        """
        self.log.warning(message, extra={'short_message': message, 'level': 4,
                                         'ml_service': self.ml_service,
                                         'service_level': self.service_level})


class JsonServiceLogger(ServiceLoggerInterface):
    class JsonFormatter(jsonlogger.JsonFormatter):
        def parse(self):
            return [
                'host',
                'short_message',
                'timestamp',
                'level',
                'service',
                'ml_service',
                'service_level',
                'ml_input',
                'ml_output',
            ]

        def add_fields(self, log_record, record, message_dict):
            super().add_fields(log_record, record, message_dict)
            log_record['host'] = gethostname()
            log_record['timestamp'] = int(time.time() * 1000) / 1000
            log_record['service'] = 'rekcurd'

    def __init__(self,
                 logger_name: str = 'rekcurd.service',
                 log_level: int = None,
                 config: RekcurdConfig = RekcurdConfig()):
        """
        Constructor
        :param logger_name:
        :param log_level:
        :param config: RekcurdConfig
        """
        self.logger_name = logger_name
        self.log_level = int(log_level or logging.DEBUG)
        self.config = config
        self.ml_service = config.APPLICATION_NAME
        self.service_level = config.SERVICE_LEVEL
        self.log = logging.getLogger(logger_name)
        handler = logging.StreamHandler()
        formatter = self.JsonFormatter()
        handler.setFormatter(formatter)
        handler.setLevel(self.log_level)
        self.log.addHandler(handler)
        self.log.setLevel(self.log_level)

    def emit(self, request, response, suppress_log_inout: bool = False) -> None:
        """
        emits service log
        """
        try:
            if suppress_log_inout:
                ml_input = ''
                ml_output = ''
            else:
                ml_input = super().to_str_from_request(request)
                ml_output = super().to_str_from_response(response)

            message = "prediction result."
            self.log.info(message, extra={'short_message': message,
                                          'level': 6, 'ml_service': self.ml_service,
                                          'service_level': self.service_level,
                                          'ml_input': ml_input,
                                          'ml_output': ml_output})
        except Exception:
            try:
                JsonSystemLogger(self.logger_name, self.log_level, self.config).exception("can't write log")
            except:
                pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Example is available on https://github.com/rekcurd/rekcurd-example/tree/master/python/sklearn-digits


from typing import Generator

from rekcurd import Rekcurd
from rekcurd.utils import PredictInput, PredictResult, EvaluateResult, EvaluateDetail, EvaluateResultDetail


class RekcurdAppTemplateApp(Rekcurd):
    def load_model(self, filepath: str) -> object:
        """ override
        TODO: Implement "load_model"
        :param filepath: ML model file path. str
        :return predictor: Your ML predictor object. object
        """
        pass

    def predict(self, predictor: object, idata: PredictInput, option: dict = None) -> PredictResult:
        """ override
        TODO: Implement "predict"
        :param predictor: Your ML predictor object. object
        :param idata: Input data. PredictInput, one of string/bytes/arr[int]/arr[float]/arr[string]
        :param option: Miscellaneous. dict
        :return result: Result. PredictResult
            result.label: Label. One of string/bytes/arr[int]/arr[float]/arr[string]
            result.score: Score. One of float/arr[float]
            result.option: Miscellaneous. dict
        """
        pass

    def evaluate(self, predictor: object, filepath: str) -> Generator[EvaluateResultDetail, None, EvaluateResult]:
        """ override
        TODO: Implement "evaluate"
        :param predictor: Your ML predictor object. object
        :param filepath: Evaluation data file path. str
        :return result: Result. EvaluateResult
            result.num: Number of data. int
            result.accuracy: Accuracy. float
            result.precision: Precision. arr[float]
            result.recall: Recall. arr[float]
            result.fvalue: F1 value. arr[float]
            result.option: Optional metrics. dict[str, float]
        :return detail[]: Detail result of each prediction. List[EvaluateResultDetail]
            detail[].result: Prediction result. PredictResult
            detail[].is_correct: Prediction result is correct or not. bool
        """
        pass

    def get_evaluate_detail(self, filepath: str, details: Generator[EvaluateResultDetail, None, None]) -> Generator[EvaluateDetail, None, None]:
        """ override
        TODO: Implement "get_evaluate_detail"
        :param filepath: Evaluation data file path. str
        :param details: Detail result of each prediction. List[EvaluateResultDetail]
        :return rtn: Return results. Generator[EvaluateDetail, None, None]
            rtn.input: Input data. PredictInput, one of string/bytes/arr[int]/arr[float]/arr[string]
            rtn.label: Predict label. PredictLabel, one of string/bytes/arr[int]/arr[float]/arr[string]
            rtn.result: Predict detail. EvaluateResultDetail
        """
        pass


if __name__ == '__main__':
    app = RekcurdAppTemplateApp()
    app.load_config_file("./settings.yml")
    app.run()



#!/usr/bin/python
# -*- coding: utf-8 -*-


from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import Generator

from rekcurd.utils import RekcurdConfig, PredictInput, PredictResult, EvaluateResult, EvaluateDetail, EvaluateResultDetail
from rekcurd.logger import SystemLoggerInterface, ServiceLoggerInterface, JsonSystemLogger, JsonServiceLogger
from rekcurd.data_servers import DataServer


class Rekcurd(metaclass=ABCMeta):
    """
    Rekcurd
    """
    _system_logger: SystemLoggerInterface = None
    _service_logger: ServiceLoggerInterface = None
    config: RekcurdConfig = None
    data_server: DataServer = None

    @abstractmethod
    def load_model(self, filepath: str) -> object:
        """
        load_model
        :param filepath: ML model file path. str
        :return predictor: Your ML predictor object. object
        """
        raise NotImplemented()

    @abstractmethod
    def predict(self, predictor: object, idata: PredictInput, option: dict = None) -> PredictResult:
        """
        predict
        :param predictor: Your ML predictor object. object
        :param idata: Input data. PredictInput, one of string/bytes/arr[int]/arr[float]/arr[string]
        :param option: Miscellaneous. dict
        :return result: Result. PredictResult
            result.label: Label. One of string/bytes/arr[int]/arr[float]/arr[string]
            result.score: Score. One of float/arr[float]
            result.option: Miscellaneous. dict
        """
        raise NotImplemented()

    @abstractmethod
    def evaluate(self, predictor: object, filepath: str) -> Generator[EvaluateResultDetail, None, EvaluateResult]:
        """
        evaluate
        :param predictor: Your ML predictor object. object
        :param filepath: Evaluation data file path. str
        :return result: Result. EvaluateResult
            result.num: Number of data. int
            result.accuracy: Accuracy. float
            result.precision: Precision. arr[float]
            result.recall: Recall. arr[float]
            result.fvalue: F1 value. arr[float]
            result.option: Optional metrics. dict[str, float]
        :generate detail[]: Detail result of each prediction. List[EvaluateResultDetail]
            detail[].result: Prediction result. PredictResult
            detail[].is_correct: Prediction result is correct or not. bool
        """
        raise NotImplemented()

    @abstractmethod
    def get_evaluate_detail(self, filepath: str, details: Generator[EvaluateResultDetail, None, None]) -> Generator[EvaluateDetail, None, None]:
        """
        get_evaluate_detail
        :param filepath: Evaluation data file path. str
        :param details: Detail result of each prediction. Generator[EvaluateResultDetail, None, None]
        :return rtn: Return results. Generator[EvaluateDetail, None, None]
            rtn.input: Input data. PredictInput, one of string/bytes/arr[int]/arr[float]/arr[string]
            rtn.label: Predict label. PredictLabel, one of string/bytes/arr[int]/arr[float]/arr[string]
            rtn.result: Predict detail. EvaluateResultDetail
        """
        raise NotImplemented()

    @property
    def system_logger(self):
        return self._system_logger

    @system_logger.setter
    def system_logger(self, system_logger: SystemLoggerInterface):
        if isinstance(system_logger, SystemLoggerInterface):
            self._system_logger = system_logger
        else:
            raise TypeError("Invalid system_logger type.")

    @property
    def service_logger(self):
        return self._service_logger

    @service_logger.setter
    def service_logger(self, service_logger: ServiceLoggerInterface):
        if isinstance(service_logger, ServiceLoggerInterface):
            self._service_logger = service_logger
        else:
            raise TypeError("Invalid service_logger type.")

    def load_config_file(self, config_file: str):
        self.config = RekcurdConfig(config_file)

    def run(self, host: str = None, port: int = None, max_workers: int = None, **options):
        import grpc
        import os
        import time
        from concurrent import futures
        from rekcurd.protobuf import rekcurd_pb2_grpc
        from rekcurd import RekcurdDashboardServicer, RekcurdWorkerServicer

        if self.config is None:
            self.config = RekcurdConfig()
        if host and "service_insecure_host" in options:
            options["service_insecure_host"] = host
        if port and "service_insecure_port" in options:
            options["service_insecure_port"] = port
        self.config.set_configurations(**options)

        self.data_server = DataServer(self.config)
        if self._system_logger is None:
            self._system_logger = JsonSystemLogger(config=self.config)
        if self._service_logger is None:
            self._service_logger = JsonServiceLogger(config=self.config)
        self.system_logger.info("Service start.")
        _host = "127.0.0.1"
        _port = 5000
        _max_workers = 1
        host = host or self.config.SERVICE_INSECURE_HOST or _host
        port = int(port or self.config.SERVICE_INSECURE_PORT or _port)
        max_workers = int(max_workers or _max_workers)

        try:
            self.system_logger.info("Download model.")
            model_path = self.data_server.get_model_path()
            self.system_logger.info("Initialize predictor.")
            predictor = self.load_model(model_path)
            if predictor is None:
                raise Exception("Error: No predictor found. Need your \"Rekcurd\" implementation.")
        except Exception as e:
            self.system_logger.error(str(e))
            print(str(e))
            return

        rekcurd_pack = RekcurdPack(self, predictor)
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
        rekcurd_pb2_grpc.add_RekcurdDashboardServicer_to_server(
            RekcurdDashboardServicer(rekcurd_pack), server)
        rekcurd_pb2_grpc.add_RekcurdWorkerServicer_to_server(
            RekcurdWorkerServicer(rekcurd_pack), server)
        server.add_insecure_port("{0}:{1}".format(host, port))
        self.system_logger.info("Start rekcurd worker on {0}:{1}".format(host, port))
        server.start()
        try:
            while os.getenv("REKCURD_UNITTEST", "False").lower() == 'false':
                time.sleep(86400)
        except KeyboardInterrupt:
            self.system_logger.info("Shutdown rekcurd worker.")
        finally:
            server.stop(0)

    # TODO: DEPRECATED BELOW
    __type_input = None
    __type_output = None

    def set_type(self, type_input: Enum, type_output: Enum) -> None:
        self.__type_input = type_input
        self.__type_output = type_output

    def get_type_input(self) -> Enum:
        return self.__type_input

    def get_type_output(self) -> Enum:
        return self.__type_output


class RekcurdPack:
    def __init__(self, app: Rekcurd, predictor: object):
        self.app = app
        self.predictor = predictor



# coding: utf-8


import pickle

from pathlib import Path
from typing import Iterator, Generator

from rekcurd.protobuf import rekcurd_pb2
from rekcurd.utils import RekcurdConfig, ModelModeEnum, EvaluateResultDetail, EvaluateResult
from .data_handler import convert_to_valid_path
from .local_handler import LocalHandler
from .ceph_handler import CephHandler
from .aws_s3_handler import AwsS3Handler
from .gcs_handler import GcsHandler


class DataServer(object):
    """DataServer
    Return file path. Download files if needed.
    """

    def __init__(self, config: RekcurdConfig):
        self.config = config
        if config.MODEL_MODE_ENUM == ModelModeEnum.LOCAL:
            self._api_handler = LocalHandler(config)
        elif config.MODEL_MODE_ENUM == ModelModeEnum.CEPH_S3:
            self._api_handler = CephHandler(config)
        elif config.MODEL_MODE_ENUM == ModelModeEnum.AWS_S3:
            self._api_handler = AwsS3Handler(config)
        elif config.MODEL_MODE_ENUM == ModelModeEnum.GCS:
            self._api_handler = GcsHandler(config)
        else:
            raise ValueError("Invalid ModelModeEnum value.")

    def get_model_path(self) -> str:
        local_filepath = Path(self._api_handler.LOCAL_MODEL_DIR, self._api_handler.MODEL_FILE_NAME)
        if not local_filepath.exists():
            local_filepath.parent.mkdir(parents=True, exist_ok=True)
            self._api_handler.download(self.config.MODEL_FILE_PATH, str(local_filepath))
        return str(local_filepath)

    def switch_model(self, filepath: str) -> str:
        valid_path = convert_to_valid_path(filepath)
        if filepath != str(valid_path):
            raise Exception(f'Error: Invalid file path specified -> {filepath}')

        local_filepath = Path(self._api_handler.LOCAL_MODEL_DIR, valid_path.name)
        if not local_filepath.exists():
            local_filepath.parent.mkdir(parents=True, exist_ok=True)
            self._api_handler.download(filepath, str(local_filepath))
        self._api_handler.MODEL_FILE_NAME = valid_path.name
        return str(local_filepath)

    def upload_model(self, request_iterator: Iterator[rekcurd_pb2.UploadModelRequest]) -> str:
        first_req = next(request_iterator)
        filepath = first_req.path
        valid_path = convert_to_valid_path(filepath)
        if filepath != str(valid_path):
            raise Exception(f'Error: Invalid file path specified -> {filepath}')

        local_filepath = Path(self._api_handler.LOCAL_MODEL_DIR, valid_path.name)
        local_filepath.parent.mkdir(parents=True, exist_ok=True)
        with local_filepath.open(mode='wb') as f:
            f.write(first_req.data)
            for request in request_iterator:
                f.write(request.data)
            del first_req
        self._api_handler.upload(filepath, str(local_filepath))
        return str(local_filepath)

    def get_evaluation_data_path(self, filepath: str) -> str:
        valid_path = convert_to_valid_path(filepath)
        if filepath != str(valid_path):
            raise Exception(f'Error: Invalid file path specified -> {filepath}')

        local_filepath = Path(self._api_handler.LOCAL_EVAL_DIR, valid_path.name)
        if not local_filepath.exists():
            local_filepath.parent.mkdir(parents=True, exist_ok=True)
            self._api_handler.download(filepath, str(local_filepath))
        return str(local_filepath)

    def get_eval_result_detail(self, filepath: str) -> str:
        valid_path = convert_to_valid_path(filepath)
        if filepath != str(valid_path):
            raise Exception(f'Error: Invalid file path specified -> {filepath}')

        local_filepath = Path(self._api_handler.LOCAL_EVAL_DIR, valid_path.name)
        if not local_filepath.exists():
            local_filepath.parent.mkdir(parents=True, exist_ok=True)
            self._api_handler.download(filepath, str(local_filepath))
        return str(local_filepath)

    def upload_evaluation_data(self, request_iterator: Iterator[rekcurd_pb2.UploadEvaluationDataRequest]) -> str:
        first_req = next(request_iterator)
        filepath = first_req.data_path
        valid_path = convert_to_valid_path(filepath)
        if filepath != str(valid_path):
            raise Exception(f'Error: Invalid evaluation file path specified -> {filepath}')

        local_filepath = Path(self._api_handler.LOCAL_EVAL_DIR, valid_path.name)
        local_filepath.parent.mkdir(parents=True, exist_ok=True)
        with local_filepath.open(mode='wb') as f:
            f.write(first_req.data)
            for request in request_iterator:
                f.write(request.data)
            del first_req
        self._api_handler.upload(filepath, str(local_filepath))
        return str(local_filepath)

    def upload_evaluation_result(self, data_gen: Generator[EvaluateResultDetail, None, EvaluateResult], filepath: str) -> EvaluateResult:
        valid_path = convert_to_valid_path(filepath)
        if filepath != str(valid_path):
            raise Exception(f'Error: Invalid evaluation result file path specified -> {filepath}')

        local_filepath = Path(self._api_handler.LOCAL_EVAL_DIR, valid_path.name)
        local_filepath.parent.mkdir(parents=True, exist_ok=True)

        with local_filepath.open(mode='wb') as detail_file:
            try:
                while True:
                    pickle.dump(next(data_gen), detail_file)
            except StopIteration as e:
                evaluate_result = e.value
        self._api_handler.upload(filepath, str(local_filepath))

        return evaluate_result



# coding: utf-8


from pathlib import Path

from .data_handler import DataHandler
from rekcurd.utils import RekcurdConfig


class LocalHandler(DataHandler):
    """LocalHandler
    """
    def __init__(self, config: RekcurdConfig):
        super(LocalHandler, self).__init__(config)
        # Overwrite: No additional local directory is needed.
        self.LOCAL_MODEL_DIR = str(Path(config.MODEL_FILE_PATH).parent)

    def download(self, remote_filepath: str, local_filepath: str) -> None:
        raise Exception(
            "Error: No such a file. Local mode requires to store"
            " a ML model file on the directory you specified. "
            "-> \"{}\"".format(local_filepath))

    def upload(self, remote_filepath: str, local_filepath: str) -> None:
        pass
